%% Machine Learning Online Class
%  Exercise 8 | Anomaly Detection and Collaborative Filtering
%
%  Instructions
%  ------------
%
%  This file contains code that helps you get started on the
%  exercise. You will need to complete the following functions:
%
%     cofiCostFunc.m
%
%  For this exercise, you will not need to change any code in this file,
%  or any other files other than those mentioned above.
%

%% =============== Part 1: Loading movie ratings dataset ================
%  You will start by loading the movie ratings dataset to understand the
%  structure of the data.
%  
fprintf('Loading movie ratings dataset.\n\n');

%  Load data
load('movies.mat');
movieList = loadMovieList();

%  Y is a 1682x943 matrix, containing ratings (1-5) of 1682 movies on 
%  943 users
%
%  R is a 1682x943 matrix, where R(i,j) = 1 if and only if user j gave a
%  rating to movie i

%  From the matrix, we can compute statistics like average rating.
fprintf('Average rating for movie 1 (Toy Story): %f / 5\n\n', ...
        mean(Y(1, R(1, :))));

%  We can "visualize" the ratings matrix by plotting it with imagesc
imagesc(Y);
ylabel('Movies');
xlabel('Users');

fprintf('\nProgram paused. Press enter to continue.\n');
pause;
%{
%% ============ Part 2: Collaborative Filtering Cost Function ===========
%  You will now implement the cost function for collaborative filtering.
%  To help you debug your cost function, we have included set of weights
%  that we trained on that. Specifically, you should complete the code in 
%  cofiCostFunc.m to return J.

%  Load pre-trained weights (X, Theta, num_users, num_movies, num_features)
load ('movieParams.mat');

%  Reduce the data set size so that this runs faster
num_users = 4; num_movies = 5; num_features = 3;
X = X(1:num_movies, 1:num_features);
Theta = Theta(1:num_users, 1:num_features);
Y = Y(1:num_movies, 1:num_users);
R = R(1:num_movies, 1:num_users);

%  Evaluate cost function
J = cofiCostFunc([X(:) ; Theta(:)], Y, R, num_users, num_movies, ...
               num_features, 0);
           
fprintf(['Cost at loaded parameters: %f '...
         '\n(this value should be about 22.22)\n'], J);

fprintf('\nProgram paused. Press enter to continue.\n');
pause;


%% ============== Part 3: Collaborative Filtering Gradient ==============
%  Once your cost function matches up with ours, you should now implement 
%  the collaborative filtering gradient function. Specifically, you should 
%  complete the code in cofiCostFunc.m to return the grad argument.
%  
fprintf('\nChecking Gradients (without regularization) ... \n');

%  Check gradients by running checkCostFunction
checkCostFunction;

fprintf('\nProgram paused. Press enter to continue.\n');
pause;

%% ========= Part 4: Collaborative Filtering Cost Regularization ========
%  Now, you should implement regularization for the cost function for 
%  collaborative filtering. You can implement it by adding the cost of
%  regularization to the original cost computation.
%  

%  Evaluate cost function
J = cofiCostFunc([X(:) ; Theta(:)], Y, R, num_users, num_movies, ...
               num_features, 1.5);
           
fprintf(['Cost at loaded parameters (lambda = 1.5): %f '...
         '\n(this value should be about 31.34)\n'], J);

fprintf('\nProgram paused. Press enter to continue.\n');
pause;


%% ======= Part 5: Collaborative Filtering Gradient Regularization ======
%  Once your cost matches up with ours, you should proceed to implement 
%  regularization for the gradient. 
%

% 
fprintf('\nChecking Gradients (with regularization) ... \n');

%  Check gradients by running checkCostFunction
checkCostFunction(1.5);

fprintf('\nProgram paused. Press enter to continue.\n');
pause;
%}

%% ============== Part 6: Entering ratings for a new user ===============
%  Before we will train the collaborative filtering model, we will first
%  add ratings that correspond to a new user that we just observed. This
%  part of the code will also allow you to put in your own ratings for the
%  movies in our dataset!
%

num_movies = 1682;
%  Initialize my ratings
my_ratings = zeros(num_movies, 1);

% Check the file movie_idx.txt for id of each movie in our dataset
% For example, Toy Story (1995) has ID 1, so to rate it "4", you can set
%my_ratings(1) = 4;

possible_movies=1:num_movies;
fprintf('Hint: If you are fed up with rating movies, you can hit the button ''q'' to quit.');
fprintf('\n\nNew user ratings:\n');
while (nnz(my_ratings) < 15)
  choosable_movie_num = length(possible_movies);
  random_movie_id = unidrnd(choosable_movie_num - 1); % this minus one is kinda ugly there; but it makes life easier so it's kinda OK there
  possible_movies = [possible_movies(1 : random_movie_id - 1) possible_movies(random_movie_id + 1 : choosable_movie_num)];
  hint =   sprintf('Don''t worry, you only have to provide %d more ratings.\n', 15 - nnz(my_ratings));
  prompt = strcat('Please rate movie ''',  movieList{random_movie_id}, ''' from 0 to 5.\n', hint);
  answer = input(sprintf(prompt),'s');

  while (length(answer) > 0 && answer ~= 'q' && (length(answer) > 1 || answer(1) < '0' || answer(1) > '5'))
    answer = input(sprintf('Did not I mention that the ratings should be in the range [0, 5] (exclusively)?\nSo please provide another rating.\n'), 's');
  end

  if (answer == 'q')
   break;
  end
  % if we are here we can be pretty sure that the answer is either the empty string of one of the digits in the range [0,5]
  if (length(answer) > 0 && answer(1) ~= '0')
    my_ratings(random_movie_id) = str2num(answer(1));
  else
    disp(sprintf('You decided not to rate this movie.\n'))
  end
end


%fprintf('\nProgram paused. Press enter to continue.\n');
%pause;


%% ================== Part 7: Learning Movie Ratings ====================
%  Now, you will train the collaborative filtering model on a movie rating 
%  dataset of 1682 movies and 943 users
%

%  Y is a 1682x943 matrix, containing ratings (1-5) of 1682 movies by 
%  943 users
%
%  R is a 1682x943 matrix, where R(i,j) = 1 if and only if user j gave a
%  rating to movie i

%  Add our own ratings to the data matrix and normalize ratings
fprintf('\nNormalizing the ratings...\n');
Y = [my_ratings Y];
R = [(my_ratings ~= 0) R];
[Ynorm, Ymean] = normalizeRatings(Y, R);

%num_latent_factors = 100; %in case we wish more latent dimensions
num_latent_factors_for_visualisation = 2;
fprintf('\nComputing the SVD decomposition of the data...\n');
[U, S, V] = svd(Ynorm);
Ui = U(:, 1:num_latent_factors_for_visualisation);
Si = S(1:num_latent_factors_for_visualisation, 1:num_latent_factors_for_visualisation);
Vi = V(:, 1:num_latent_factors_for_visualisation);

Ynorm_2d = Ui * Si * Vi';
latent2d_movie_space = Ui * sqrt(Si);
printf('The reconstruction had an error of %f', norm(Ynorm - Ynorm_2d));
scatter(latent2d_movie_space(:,1), latent2d_movie_space(:,2));
labels = num2str((1 : size(latent2d_movie_space, 1))','%d');
text(latent2d_movie_space(:,1), latent2d_movie_space(:,2), labels)
fprintf('\nProgram paused. Press enter to continue.\n');
pause;

%  Useful Values
num_users = size(Y, 2);
num_movies = size(Y, 1);

num_features = 50;

fprintf('\nTraining collaborative filtering...\n');

% Set Initial Parameters (Theta, X)
X = randn(num_movies, num_features);
Theta = randn(num_users, num_features);

initial_parameters = [X(:); Theta(:)];

% Set options for fmincg
options = optimset('GradObj', 'on', 'MaxIter', 150);

% Set Regularization
lambda = 10;
theta = fmincg (@(t)(cofiCostFunc(t, Y, R, num_users, num_movies, num_features, lambda)), ...
                initial_parameters, options);

% Unfold the returned theta back into U and W
X = reshape(theta(1 : num_movies * num_features), num_movies, num_features);
Theta = reshape(theta(num_movies * num_features + 1 : end), num_users, num_features);

fprintf('Recommender system learning completed.\n');

fprintf('\nProgram paused. Press enter to continue.\n');
pause;

%% ================== Part 8: Recommendation for you ====================
%  After training the model, you can now make recommendations by computing
%  the predictions matrix.
%

p = X * Theta';
my_predictions = p(:, 1) + Ymean;

[r, ix] = sort(my_predictions, 'descend');
fprintf('\nTop recommendations for you:\n');
for i=1:10
    j = ix(i);
    fprintf('Predicting rating %.2f for movie %s\n', my_predictions(j), movieList{j});
end

fprintf('\n\nOriginal ratings provided:\n');
for i = 1:length(my_ratings)
    if my_ratings(i) > 0 
        fprintf('Rated %d for %s\n', my_ratings(i), movieList{i});
    end
end
